from typing import TYPE_CHECKING
from mikro_next.scalars import ArrayLike, MeshLike, ParquetLike, FileLike
import asyncio
import s3fs
from aiobotocore.session import get_session
import botocore
from concurrent.futures import ThreadPoolExecutor

from .errors import PermissionsError, UploadError
from zarr.storage import FsspecStore
import zarr
import zarr.api.asynchronous as async_api
import aiohttp

if TYPE_CHECKING:
    from mikro_next.api.schema import Credentials, PresignedPostCredentials
    from mikro_next.datalayer import DataLayer


def _store_xarray_input(
    xarray: ArrayLike,
    credentials: "Credentials",
    endpoint_url: "DataLayer",
) -> str:
    """Stores an xarray in the DataLayer"""

    filesystem = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=True
    )

    # random_uuid = uuid.uuid4()
    # s3_path = f"zarr/{random_uuid}.zarr"

    array = xarray.value.transpose("c", "t", "z", "y", "x")


    s3_path = f"{credentials.bucket}/{credentials.key}"
    store = FsspecStore(filesystem, read_only=False, path=s3_path )



    try:
        zarr.save_array(store, array.to_numpy(), zarr_version=3)
        return credentials.store
    except Exception as e:
        raise UploadError(f"Error while uploading to {s3_path}") from e


async def astore_xarray_input(
    xarray: ArrayLike,
    credentials: "Credentials",
    endpoint_url: "DataLayer",
) -> str:
    """Stores an xarray in the DataLayer"""

    filesystem = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=True
    )

    # random_uuid = uuid.uuid4()
    # s3_path = f"zarr/{random_uuid}.zarr"

    array = xarray.value.transpose("c", "t", "z", "y", "x")


    s3_path = f"{credentials.bucket}/{credentials.key}"
    store = FsspecStore(filesystem, read_only=False, path=s3_path )



    try:
        await async_api.save_array(store, array.to_numpy(), zarr_version=3)
        return credentials.store
    except Exception as e:
        raise UploadError(f"Error while uploading to {s3_path}") from e


def _store_parquet_input(
    parquet_input: ParquetLike,
    credentials: "Credentials",
    endpoint_url: str,
) -> str:
    """Stores an xarray in the DataLayer"""
    import pyarrow.parquet as pq
    from pyarrow import Table
    import aiohttp

    filesystem = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
    )

    table: Table = Table.from_pandas(parquet_input.value)

    try:
        s3_path = f"s3://{credentials.bucket}/{credentials.key}"
        pq.write_table(table, s3_path, filesystem=filesystem)
        return credentials.store
    except Exception as e:
        raise UploadError(f"Error while uploading to {s3_path}") from e


async def astore_mesh_file(
    mesh: MeshLike,
    credentials: "PresignedPostCredentials",
    datalayer: "DataLayer",
    endpoint_url: str,
):
    
    
    endpoint_url = await datalayer.get_endpoint_url()
    
    async with aiohttp.ClientSession() as session:
        form_data = aiohttp.FormData()
        form_data.add_field("key", credentials.key)
        form_data.add_field("policy", credentials.policy)
        form_data.add_field("x-amz-algorithm", credentials.x_amz_algorithm)
        form_data.add_field("x-amz-credential", credentials.x_amz_credential)
        form_data.add_field("x-amz-date", credentials.x_amz_date)
        form_data.add_field("x-amz-signature", credentials.x_amz_signature)
        form_data.add_field("file", mesh.value, filename="mesh.obj", content_type="application/octet-stream")

        url = endpoint_url + "/" + credentials.bucket

        async with session.post(url, data=form_data) as resp:
            if resp.status not in {200, 204}:
                body = await resp.text()
                raise UploadError(f"Error while uploading mesh: HTTP {resp.status}: {body}")
    
    
    return credentials.store



async def aupload_bigfile(
    file: FileLike,
    credentials: "Credentials",
    datalayer: "DataLayer",
    executor: ThreadPoolExecutor = None,
) -> str:
    """Store a DataFrame in the DataLayer"""
    session = get_session()

    endpoint_url = await datalayer.get_endpoint_url()

    async with session.create_client(
        "s3",
        region_name="us-west-2",
        endpoint_url=endpoint_url,
        aws_secret_access_key=credentials.secret_key,
        aws_access_key_id=credentials.access_key,
        aws_session_token=credentials.session_token,
    ) as client:
        try:
            await client.put_object(
                Bucket=credentials.bucket, Key=credentials.key, Body=file.value
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "InvalidAccessKeyId":
                return PermissionsError(
                    "Access Key is invalid, trying to get new credentials"
                )

            raise e

    return credentials.store


async def aupload_xarray(
    array: ArrayLike,
    credentials: "Credentials",
    datalayer: "DataLayer",
    executor: ThreadPoolExecutor,
) -> str:
    """Store a DataFrame in the DataLayer"""
    return await astore_xarray_input(array, credentials, await datalayer.get_endpoint_url())


async def aupload_parquet(
    parquet: ParquetLike,
    credentials: "Credentials",
    datalayer: "DataLayer",
    executor: ThreadPoolExecutor,
) -> str:
    """Store a DataFrame in the DataLayer"""
    co_future = executor.submit(
        _store_parquet_input, parquet, credentials, await datalayer.get_endpoint_url()
    )
    return await asyncio.wrap_future(co_future)



from .mikro_next import MikroNext
from .utils import v, e, m, rm, rechunk
try:
    from .arkitekt import MikroService
except ImportError:
    pass
try:
    from .rekuest import structure_reg
    print("Imported structure_reg")
except ImportError as e:
    print("Could not import structure_reg", e)
    pass


__all__ = [
    "MikroNext",
    "v",
    "e",
    "m",
    "rm",
    "rechunk",
    "structure_reg",
    "MikroService",
]




from rekuest_next.structures.default import (
    get_default_structure_registry,
    id_shrink,
)
from rekuest_next.api.schema import PortScope
from rekuest_next.widgets import SearchWidget
from mikro_next.api.schema import (
    Image,
    aget_image,
    SearchImagesQuery,
    Dataset,
    Stage,
    aget_stage,
    File,
    Table,
    aget_file,
    SearchStagesQuery,
    SearchTablesQuery,
    SearchFilesQuery,
    aget_rgb_context,
    RGBContext,
    aget_dataset,
    aget_table,
)
from mikro_next.api.schema import (
    Snapshot,
    aget_snapshot,
    SearchSnapshotsQuery,
    ROI,
    aget_roi,
    aget_rendered_plot,
    RenderedPlot,
    SearchRoisQuery,
    SearchRenderedPlotsQuery,
    Mesh,
    aget_mesh,
    SearchMeshesQuery,
)

structure_reg = get_default_structure_registry()

structure_reg.register_as_structure(
    Image,
    identifier="@mikro/image",
    aexpand=aget_image,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(
        query=SearchImagesQuery.Meta.document, ward="mikro"
    ),
)
structure_reg.register_as_structure(
    Snapshot,
    identifier="@mikro/snapshot",
    aexpand=aget_snapshot,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(
        query=SearchSnapshotsQuery.Meta.document, ward="mikro"
    ),
)

structure_reg.register_as_structure(
    ROI,
    identifier="@mikro/roi",
    aexpand=aget_roi,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(query=SearchRoisQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    Stage,
    identifier="@mikro/stage",
    aexpand=aget_stage,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(
        query=SearchStagesQuery.Meta.document, ward="mikro"
    ),
)
structure_reg.register_as_structure(
    Dataset,
    identifier="@mikro/dataset",
    aexpand=aget_dataset,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(
        query=SearchImagesQuery.Meta.document, ward="mikro"
    ),
)
structure_reg.register_as_structure(
    File,
    identifier="@mikro/file",
    aexpand=aget_file,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(query=SearchFilesQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    RGBContext,
    identifier="@mikro/rbgcontext",
    aexpand=aget_rgb_context,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
)

structure_reg.register_as_structure(
    RenderedPlot,
    identifier="@mikro/renderedplot",
    aexpand=aget_rendered_plot,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(
        query=SearchRenderedPlotsQuery.Meta.document, ward="mikro"
    ),
)

structure_reg.register_as_structure(
    Mesh,
    identifier="@mikro/mesh",
    aexpand=aget_mesh,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(query=SearchMeshesQuery.Meta.document, ward="mikro"),
)

structure_reg.register_as_structure(
    Table,
    identifier="@mikro/table",
    aexpand=aget_table,
    ashrink=id_shrink,
    scope=PortScope.GLOBAL,
    default_widget=SearchWidget(query=SearchTablesQuery.Meta.document, ward="mikro"),
)



import json
import os
from pydantic import Field
from rath.links.file import FileExtraction
from rath.links.dictinglink import DictingLink
from rath.links.auth import AuthTokenLink
from rath.links.split import SplitLink
from fakts_next import Fakts
from herre_next import Herre
from arkitekt_next.service_registry import BaseArkitektService, Params
from arkitekt_next.base_models import Requirement

from mikro_next.mikro_next import MikroNext
from mikro_next.rath import MikroNextLinkComposition, MikroNextRath
from rath.links.split import SplitLink
from fakts_next.contrib.rath.aiohttp import FaktsAIOHttpLink
from fakts_next.contrib.rath.graphql_ws import FaktsGraphQLWSLink
from herre_next.contrib.rath.auth_link import HerreAuthLink
from mikro_next.contrib.fakts.datalayer import FaktsDataLayer
from mikro_next.links.upload import UploadLink
from mikro_next.datalayer import DataLayer
from graphql import OperationType
from herre_next import Herre
from fakts_next import Fakts
from arkitekt_next.service_registry import (
    BaseArkitektService,
    Params,
    get_default_service_registry,
)

from arkitekt_next.base_models import Manifest

try:
    from rekuest_next.links.context import ContextLink
    from rath.links.compose import TypedComposedLink

    class ArkitektMikroNextLinkComposition(TypedComposedLink):
        fileextraction: FileExtraction = Field(default_factory=FileExtraction)
        """ A link that extracts files from the request and follows the graphql multipart request spec"""
        dicting: DictingLink = Field(default_factory=DictingLink)
        """ A link that converts basemodels to dicts"""
        upload: UploadLink
        """ A link that uploads supported data types like numpy arrays and parquet files to the datalayer"""
        auth: AuthTokenLink
        """ A link that adds the auth token to the request"""
        """ A link that splits the request into a http and a websocket request"""
        assignation: ContextLink = Field(default_factory=ContextLink)
        split: SplitLink

except ImportError:
    ArkitektMikroNextLinkComposition = MikroNextLinkComposition


class ArkitektMikroNextRath(MikroNextRath):
    link: ArkitektMikroNextLinkComposition


class ArkitektNextMikroNext(MikroNext):
    rath: ArkitektMikroNextRath
    datalayer: DataLayer


def build_relative_path(*path: str) -> str:
    return os.path.join(os.path.dirname(__file__), *path)


class MikroService(BaseArkitektService):

    def get_service_name(self):
        return "mikro"

    def build_service(
        self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest
    ):
        datalayer = FaktsDataLayer(fakts_group="datalayer", fakts=fakts)

        return ArkitektNextMikroNext(
            rath=ArkitektMikroNextRath(
                link=ArkitektMikroNextLinkComposition(
                    auth=HerreAuthLink(herre=herre),
                    upload=UploadLink(
                        datalayer=datalayer,
                    ),
                    split=SplitLink(
                        left=FaktsAIOHttpLink(
                            fakts_group="mikro", fakts=fakts, endpoint_url="FAKE_URL"
                        ),
                        right=FaktsGraphQLWSLink(
                            fakts_group="mikro", fakts=fakts, ws_endpoint_url="FAKE_URL"
                        ),
                        split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                    ),
                )
            ),
            datalayer=datalayer,
        )

    def get_requirements(self):
        return [
            Requirement(
                key="mikro",
                service="live.arkitekt.mikro",
                description="An instance of ArkitektNext Mikro to make requests to the user's data",
                optional=True,
            ),
            Requirement(
                key="datalayer",
                service="live.arkitekt.s3",
                description="An instance of ArkitektNext Datalayer to make requests to the user's data",
                optional=True,
            ),
        ]

    def get_graphql_schema(self):
        schema_graphql_path = build_relative_path("api", "schema.graphql")
        with open(schema_graphql_path) as f:
            return f.read()

    def get_turms_project(self):
        turms_prject = build_relative_path("api", "project.json")
        with open(turms_prject) as f:
            return json.loads(f.read())




get_default_service_registry().register(MikroService())